/*
 * Isbell1958.js
 *
 *  Copyright (C) 2011 Stefan Bolus, University of Kiel, Germany
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


function is_less_desir (manager, r, i, j)
{
	if (i === j) return false;
	else {
		// { S-i ; S in set(r), j not in S } \subsetneq { S-j ; S in set(r) } 
		return less_manip(r, new Without(manager, j, new Remove(manager, i)), 
				r, new Without(manager,i, new Remove(manager, j)));
	}
}


/**
 * @class Type
 * 
 * The type of a player. Always belongs to a desirability relation.
 * 
 * @ctor
 * 
 * @param rep Representative of the type. Used internally only.
 */
function Type (index, alad, rep) {
	this.index = index;
	this.alad = alad;
	this.rep = rep;
	
	/**
	 * Returns the indices of the types which are directed successors of the
	 * current type.
	 */
	this.getDirectSuccessors = function () {
		var reps = this.alad.getPreorder().getDirectSuccessingRepsOf (this.rep);
		var succs = [];
		for (var i in reps) {
			/* There are no feedback loops, because the there are no cycles in 
			 * the graph of the underlying ALAD relation. */
			succs.push(this.alad.getTypeOfPlayer(reps[i]).index);
		}
		return succs;
	};
	
	/**
	 * Returns the number of players of that type.
	 */
	this.size = function () { 
		return this.alad.getPreorder().getEquivClasses().getSetOf(this.rep).length;
	};
};


/*!
 * Isbell's at-least-as-desirable-as relation.
 * 
 * \param W The QOBDD of the winning coalitions.
 * \param not_greater_func If true for i,j player i is NOT more desirable than
 * 	 	j. If false, implies nothing. Optional, can be null or undefined.
 * \param not_equal_func If true for i,j, both player are NOT symmetric. If
 *      false implies nothing otherwise. Optional, can be null or undefined.
 * \param equal_func If true for i,j, both players ARE symmetric. If false, 
 *      implies nothing. Optional, can be null or undefined.
 */
function AtLeastAsDesirable (manager, W, not_greater_func, not_equal_func, equal_func)
{
	this.n = W.getVarCount();
	this.W = W;
	this.manager = manager;
	
	this.preorder = new Preorder (this.n, 
			function /*is_greater*/ (i,j) { 
					if (not_greater_func(i,j)) return false;
					else return is_less_desir (manager, W, j, i); },
			function /*are equal*/ (i,j) {
					if (not_equal_func(i,j)) return false;
					else {
						// { S-i ; S in set(r), j not in S } = { S-j ; S in set(r), i not in S }
						return equal_manip(W, new Without(manager, j, new Remove(manager, i)),
                            W, new Without(manager,i,new Remove(manager, j)));
					}
				},
			equal_func);
	
	
	
	/**************************************************************************
	 *                                 Operations                             *
	 *************************************************************************/
	
	this.isTotal = function () { return this.preorder.isTotal(); };
	this.getNonIncrOrder = function () { return this.preorder.getNonIncrOrder(); };
	// It is assumed by FWVGReprModelService, that if the game is consecutive, the players
	// in the i-th class precede the players in the (i+1)-th class.
	this.getSymmetryClasses = function () { return this.preorder.getEquivClasses().getSets(); };
	this.getDirectSuccessorsOf = function (i) { return this.preorder.getDirectSuccessorsOf(i); };
	
	this.areSymmetric = function (i,j) { return this.preorder.areEqual(i, j); };
	this.isMoreDesirable = function (i,j) { return this.preorder.isGreater(i, j); };
	this.isAtLeastAsDesirable = function (i,j) { return this.preorder.isGreaterEqual(i, j); };
	
	this.getNotTotalWitnesses = function () { return this.preorder.getNotTotalWitnesses(); };
	
	this.getPreorder = function () { return this.preorder; };
	
	
	this.types = [];
	this.have_types = false;
	this.rep_to_type = [];
	this.next_type_index = 0;

	this.getTypeCount = function () { return this.getSymmetryClasses().length; };
	this.getTypeOfPlayer = function (i) {
		var rep = this.getPreorder().getEquivClasses().find(i);
		if (this.rep_to_type[rep] === undefined) {
			var type = new Type (this.next_type_index++, this, rep);
			this.rep_to_type[rep] = type;
			this.types[type.index] = type;
		}
		return this.rep_to_type[rep];
	};
	this.getTypes = function () {
		if ( ! this.have_types) {
			var classes = this.getPreorder().getEquivClasses();
			var reps = classes.getReps();
			
			for (var i in reps)
				this.getTypeOfPlayer(reps[i]);
			
			this.have_types = true;
		}
		return this.types;
	};
	/**
	 * Returns an array that maps the i'th player ot the index of its type,
	 * i.e. this.getTypes()[this.getPlayerToTypeMap(i)] would return in the
	 * actual type object of player i.
	 */
	this.getPlayerToTypeMap = function () {
		var map = [];
		for (var i = 0 ; i < this.n ; ++i)
			map.push (this.getTypeOfPlayer(i).index);
		return map;
	};
	
	/**
	 * Returns a vector of vectors [A1,..,Am] where Ai contains the players
	 * of a symmetry class. The classes are not necessarily ordered.
	 */
	this.getTypeVector = function () {
		return this.getPreorder().getEquivClasses().getSets();
	};
	
	
	/**
	 * 
	 */
	this.getTypeOrder = function () {
		var preorder = this.getPreorder();
		var classes = this.getSymmetryClasses();
		return new Preorder (classes.length,
				function (i,j) { return preorder.isGreater(classes[i][0], classes[j][0]); } /*is_greater*/, 
				function (i,j) { return i === j; } /*is_equal*/);
	}
	
}
