/*
 * ModelCollectorConsecutive.js
 *
 *  Copyright (C) 2011 Stefan Bolus, University of Kiel, Germany
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


/**
 * @constructor
 * 
 * Symmetric nodes have to be consecutive for this approach.
 * 
 * @param f The QOBDD which models are to be collected.
 * @param N The equivalence classe of the players.
 * 
 * @code
	var alad = game.getAlad();
	var N = alad.getTypeVector();
	var Wmin = game.getMinWinQOBDD();
	var c = new ModelCollectorConsecutive (Wmin, N);
	c.getModels().join(" || ");
 * @encode
 * 
 * @see ModelCollectorConsecutive.getModels
 */
function ModelCollectorConsecutive (f,N) {
	this._f = f;
	this._n_types = N.length;
	this._N = N;
	
	if ( !is_consecutive(this._N))
		throw "Players in game are not consecutive."
	
	/* Prepare the map which maps players to its type index. */
	this._map = [];
    for (var t = 0 ; t < this._n_types ; ++t)
    	for (var i in N[t])
    		this._map[N[t][i]] = t;
    
    this._info_store = []; // Maps node IDs to NodeInfo.
    
    this._newModel = function (n, def) {
    	var arr = [];
    	for (var i = 0 ; i < n ; ++i)
    		arr.push(def);
    	return arr;
    };
    
    /**
     * @param v The \ref QOBDD.Node.
     * @returns Returns the models for the given node.
     */
    this._getModels = function (v) {
  	
    	if (v.isOne()) return [ this._newModel(this._n_types, 0) ];
    	else if (v.isZero()) return [];
    	else {
    		var id = v.getId();
    		var info = this._info_store[id];
    		
    		if (info === undefined) {
    			info = this._info_store[id] = { models : [] };
    			
    			var label = v.getVar();
    			var type = this._map[label];
    			
    			/* Compute v's models_by_type. */
    			this._modelsByType (v, this._N[type].length);
    		
    			for (var k in info.models_by_type) {
    				var models_k = info.models_by_type[k];

    				for (var j in models_k) {
    					var model = models_k[j].slice(0); // we need a copy.
    					model[type] = k;    				
    					info.models.push (model);
    				}
    			}
    		}
    		
    		return info.models;    	
    	}
    };
    
    
    /**
     * @param v The \ref QOBDD.Node.
     * @param left The number of symmetric player below plus the current level.
     * @returns Returns nothing.
     */
    this._modelsByType = function (v, left) {
    	var info = this._info_store[v.getId()];
    	
		if (info === undefined)
			info = this._info_store[v.getId()] = {};
		
    	if (info.models_by_type === undefined) {
    		info.models_by_type = [];
    		for (var i = 0 ; i <= left + 1 ; ++i)
    			info.models_by_type[i] = null;
    		
    		if (1 === left) {
    			info.models_by_type[1] = this._getModels(v.getThen());
    			info.models_by_type[0] = this._getModels(v.getElse());
    		}
    		else {
    			this._modelsByType (v.getThen(), left-1);
    			this._modelsByType (v.getElse(), left-1);
    			
    			var infoE = this._info_store[v.getElse().getId()];
    			for (var k = 0 ; k < left ; ++k) 
    				info.models_by_type[k] = infoE.models_by_type[k];
    			
    			info.models_by_type[left] = this._info_store[v.getThen().getId()].models_by_type[left-1];
    		}
    	}
    };
    
    
    /**
     * Computes the models for the QOBDD passed to the constructor and the
     * given equivalence class. The models are created only once and are stored
     * during the lifetime of the object. Hence, if you suppose memory
     * bottlenecks do not forget to destroy this object if no longer needed.
     * 
     * @returns Returns the models for the \ref QOBDD passed to the
     * 	constructor.
     */
    this.getModels = function () {
    	return this._getModels (this._f.getRoot());
    };
}