/*
 * ModelCounterConsecutive.js
 *
 *  Copyright (C) 2011 Stefan Bolus, University of Kiel, Germany
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


/**
 * @constructor
 * 
 * Symmetric nodes have to be consecutive for this approach.
 * 
 * @param f The QOBDD which models are to be counted.
 * @param N The equivalence classes of the players.
 * 
 * @code
	var alad = game.getAlad();
	var N = alad.getTypeVector();
	var Wmin = game.getMinWinQOBDD();
	var counter = new ModelCounterConsecutive (Wmin, N);
	counter.getModelCount();
 * @encode
 * 
 * @see ModelCounterConsecutive.getModelCount
 */
function ModelCounterConsecutive (f, N)
{
	this._f = f;
	this._n_types = N.length;
	this._N = N;
	
	if ( !is_consecutive(this._N))
		throw "Players in game are not consecutive."
	
	/* Prepare the map which maps players to its type index. */
	this._map = [];
    for (var t = 0 ; t < this._n_types ; ++t)
    	for (var i in N[t])
    		this._map[N[t][i]] = t;
    
    this._info_store = []; // Maps node IDs to NodeInfo.
    
    /**
     * 
     * @param v
     * @returns Returns the number of models in set(v).
     */
    this._first_of_type = function (v) {
    	if (v.isOne()) return 1;
    	else if (v.isZero()) return 0;
    	else {
    		var info = this._info_store[v.getId()];
    		
    		if (info === undefined) {
    			info = this._info_store[v.getId()] = { };
    			
    			info.n_models = accumulate (this._ith_of_type (v,0), 0);
    		}
    		
    		return info.n_models;
    	}
    };
    
    /**
     * 
     * @returns
     */
    this._ith_of_type = function (v,i) {
    	var info = this._info_store[v.getId()];
		
		if (info === undefined)
			info = this._info_store[v.getId()] = {};
		
		if (v.isConst())
			throw ("In ModelCounterConsecutive: v is a terminal node. ");
		
    	if (info.n_models_by_type === undefined) {
    		var type = this._map[v.getVar()];
    		var players_left = this._N[type].length - i;
    		
    		info.n_models_by_type = [];
    		for (var j = 0 ; j < players_left + 1 ; ++j)  info.n_models_by_type[j] = 0;
    		
    		if (1 === players_left) {
    			info.n_models_by_type[1] = this._first_of_type (v.getThen());
    			info.n_models_by_type[0] = this._first_of_type (v.getElse());
    		}
    		else {
    			//if (v.getElse())
    			
    			var mE = this._ith_of_type (v.getElse(), i+1);
    			
    			if (mE.length !== players_left)
    				throw (mE.length + " = mE.length !== players_left = " + players_left
    						+", type="+type + ", v.getVar()="+v.getVar()+", "
    						+"this._N[type]=["+this._N[type]+"], "
    						+"this._N[type].length="+this._N[type].length + ", "
    						+"i="+i);
    			
    			for (var k = 0 ; k < players_left ; ++k)
    				info.n_models_by_type[k] = mE[k];
    			info.n_models_by_type[players_left] = this._ith_of_type(v.getThen(), i+1)[players_left-1];
    		}
    	}
    	
    	return info.n_models_by_type;
    };
    
    this.getModelCount = function () { return this._first_of_type (this._f.getRoot()); };
}